
using System;
using System.IO;
using System.Collections.Generic;
using System.Text;

namespace mmh_ccx2vtk
{
	partial class Program
	{
		private static int VTKElemType(int Elind) {
			switch (ElementTab[Elind].TYP) {
			case 11:
				return 3;
			case 12:
				return 21;
			case 7:
				return 5;
			case 8:
				return 22;
			case 9:
				return 9;
			case 10:
				return 23;
			case 3:
				return 10;
			case 6:
				return 24;
			case 1:
				return 12;
			case 4:
				return 25;
			case 2:
			case 5:
				return 13;
			default:
				return 0;
			}
		}

		private static string VtkElementLine(int ElInd) {
			string tmpstr = "";
			int n, i;
			switch (ElementTab [ElInd].TYP) {
			case 11:
			case 12:
			case 7:
			case 8:
			case 9:
			case 10:
			case 3:
			case 6:
			case 1:

				n = ElementTab[ElInd].N.Length; tmpstr = string.Format("{0} ", n);
				for (i = 0; i < n; i++)
					tmpstr += string.Format("{0, 8}", NodeTab[ElementTab[ElInd].N[i]].vtk_nr);
				break;
			case 4:
				tmpstr = "20 ";
				for (i = 0; i < 12; i++)
					tmpstr += string.Format("{0, 8}", NodeTab[ElementTab[ElInd].N[i]].vtk_nr);
				//last eight nodes has to be renumbered
				for (i = 16; i < 20; i++)
					tmpstr += string.Format("{0, 8}", NodeTab[ElementTab[ElInd].N[i]].vtk_nr);
				for (i = 12; i < 16; i++)
					tmpstr += string.Format("{0, 8}", NodeTab[ElementTab[ElInd].N[i]].vtk_nr);
				break;
			case 2:
			case 5:
				tmpstr = "6 ";
				tmpstr += string.Format("{0, 8}", NodeTab[ElementTab[ElInd].N[0]].vtk_nr);
				tmpstr += string.Format("{0, 8}", NodeTab[ElementTab[ElInd].N[2]].vtk_nr);
				tmpstr += string.Format("{0, 8}", NodeTab[ElementTab[ElInd].N[1]].vtk_nr);
				tmpstr += string.Format("{0, 8}", NodeTab[ElementTab[ElInd].N[3]].vtk_nr);
				tmpstr += string.Format("{0, 8}", NodeTab[ElementTab[ElInd].N[5]].vtk_nr);
				tmpstr += string.Format("{0, 8}", NodeTab[ElementTab[ElInd].N[4]].vtk_nr);
				break;
			}
			return tmpstr;
		}

        private static void OutputNodResBlock(tNodResBlock nrb, int dataindex, StreamWriter writer) {
            int j, jlo, jhi;
            if (dataindex < 0) {
                jlo = 0;
                jhi = nrb.NumOfItems;
            }
            else {
                jlo = dataindex;
                jhi = dataindex + 1;
            }
            string tmpstr;
            double dblval;
            foreach (KeyValuePair<int, tNode> KVP in NodeTab) {
                if (NodeTab[KVP.Key].IsFirst) {
                    tmpstr = "";
                    for (j = jlo; j < jhi; j++) {
                        if (nrb.DICT.ContainsKey(KVP.Key))
                            dblval = nrb.DICT[KVP.Key].Data[j];
                        else
                            dblval = 0;
                        tmpstr += string.Format(NFI, "{0, 16:0.0000000E+000}", dblval);
                    }
                    writer.WriteLine(tmpstr);
                }
            }
        }

        private static void MakeVTKFile() {

			string vtk_plik = Path.GetDirectoryName(FrdFileName) +
							  Path.DirectorySeparatorChar  +
							  Path.GetFileNameWithoutExtension(FrdFileName) + "_mmh.vtk";

			string tmpstr;
			int i, j, k;

			Console.WriteLine (TXT0010, vtk_plik);

			// determine main nodes total number
			// elements ccx no.5 (15 node wedge) as they are not supported in VTK
			// has to be exported as type ccx type 2 (VTK typ 13)
			// additional nodes has to be ommited
			int nn = 0;
			foreach (KeyValuePair<int, tNode> KVP in NodeTab)
				if (KVP.Value.IsFirst)
					nn++;
			// total number of elements
			int ne = ElementTab.Length;

			using (StreamWriter writer = new StreamWriter(vtk_plik)) {
				writer.WriteLine ("# vtk DataFile Version 3.0");
				writer.WriteLine ("MMH: CalculiX output '{0}' file to 'vtk' format",
				                  Path.GetFileName(FrdFileName));
				writer.WriteLine ("ASCII");
				writer.WriteLine ("DATASET UNSTRUCTURED_GRID");
				writer.WriteLine ("POINTS {0} double", nn);

				//Nodes:
				List<int> tmpList = new List<int>(nn);
				foreach (KeyValuePair<int, tNode> KVP in NodeTab)
					if (KVP.Value.IsFirst) {
						tmpList.Add (KVP.Key);
						tmpstr = string.Format (NFI, "{0, 16:0.00000000E+000}", KVP.Value.X) +
								 string.Format (NFI, "{0, 16:0.00000000E+000}", KVP.Value.Y) +
								 string.Format (NFI, "{0, 16:0.00000000E+000}", KVP.Value.Z);
						writer.WriteLine (tmpstr);
					}

				// setting vtk numbers to nodes,
				// now they will be numbered starting from 0
				// and counted as lines with coordinates
				tNode mN;
				for (i = 0; i < tmpList.Count; i++ ) {
					mN = NodeTab [tmpList [i]];
					mN.vtk_nr = i;
					NodeTab [tmpList [i]] = mN;
				}
				writer.WriteLine (); // tmpList.Clear ();

				// new number of nodes determine
				int totn = 0;
				for (i = 0; i < ElementTab.Length; i++)
					totn += (ElementTab[i].TYP == 5 ? 6 : ElementTab [i].N.Length);

				writer.WriteLine ("CELLS {0} {1}", ne, ne  + totn);

				for (i = 0; i < ne; i++)
					writer.WriteLine (VtkElementLine(i));
				writer.WriteLine ();

				// element types:
				writer.WriteLine ("CELL_TYPES {0}", ne);
				for (i = 0; i < ne; i++)
					writer.WriteLine (VTKElemType(i));
				writer.WriteLine ();

				// cell data

				// not used in this case

				// point data
				writer.WriteLine ("POINT_DATA {0}", nn);

				//###################################################################
				//nodal results blocks iteration
				for (i = 0; i < DataTab.Count; i++) {

					if (DataTab [i].NumOfItems == 3) {
                        //the blocks with 3 items are supposed to be a vectors
                        writer.Write("VECTORS ");
                        writer.Write(DataTab[i].StepN + 1);
                        writer.Write("_");
                        writer.Write(DataTab[i].StepT.ToString(NFI));
                        writer.Write("_");
                        writer.Write(DataTab[i].Name);
                        writer.WriteLine(" double");
                        //line iteration
                        OutputNodResBlock(DataTab[i], -1, writer);

					}
                    else {
						// scalar results
						for (j = 0; j < DataTab[i].NumOfItems; j++) {
                            writer.Write("SCALARS ");
                            writer.Write(DataTab[i].StepN + 1);
                            writer.Write("_");
                            writer.Write(DataTab[i].StepT.ToString(NFI));
                            writer.Write("_");
                            writer.Write(DataTab[i].Name);
                            writer.Write("_");
                            writer.Write(DataTab[i].ItemNameTab[j]);
                            writer.WriteLine(" double");
                            writer.WriteLine ("LOOKUP_TABLE default");
                            //iteracja po liniach
                            OutputNodResBlock(DataTab[i], j, writer);
						}

					} //vector/scalar

				}  //block iteration

				//additional calculations
				// Principal stresses
				double sxx, syy, szz, sxy, syz, szx;
                double I1, I2, I3;
                double xa, xb, dxab, x;
				for (i = 0; i < DataTab.Count; i++) {
					if (DataTab[i].Name == "STRESS" || DataTab[i].Name == "TOSTRAIN" || DataTab[i].Name == "STRESSR") {
                        double[,] sprincipal = new double[NodeTab.Count,3];
                        j = -1;
                        foreach (KeyValuePair<int, tNode> KVP in NodeTab) {
                            if (NodeTab[KVP.Key].IsFirst) {
                                j++;
                                for (k = 0; k < 3; k++)
                                    sprincipal[k, 0] = 0;

                                if (DataTab[i].DICT.ContainsKey(KVP.Key)) {

                                    sxx = DataTab[i].DICT[KVP.Key].Data[0];
                                    syy = DataTab[i].DICT[KVP.Key].Data[1];
                                    szz = DataTab[i].DICT[KVP.Key].Data[2];
                                    sxy = DataTab[i].DICT[KVP.Key].Data[3];
                                    syz = DataTab[i].DICT[KVP.Key].Data[4];
                                    szx = DataTab[i].DICT[KVP.Key].Data[5];

                                    I1 = sxx + syy + szz;
                                    I2 = sxx * syy + syy * szz + szz * sxx -
                                         sxy * sxy - syz * syz - szx * szx;
                                    I3 = sxx * syy * szz -
                                         sxx * syz * syz -
                                         syy * szx * szx -
                                         szz * sxy * sxy +
                                         2 * sxy * syz * szx;

                                    xa = -(Math.Sqrt(I1 * I1 - 3 * I2) - I1) / 3;
                                    xb = +(Math.Sqrt(I1 * I1 - 3 * I2) + I1) / 3;

                                    dxab = xb - xa;

                                    x = 0;
                                    if (Poly3Root(xa - dxab, xa, I1, I2, I3, ref x))
                                        sprincipal[j, 2] = x;
                                    else
                                        Console.WriteLine("Failed to determine S3.");

                                    if (Poly3Root(xa, xb, I1, I2, I3, ref x))
                                        sprincipal[j, 1] = x;
                                    else
                                        Console.WriteLine("Failed to determine S2.");

                                    if (Poly3Root(xb, xb + dxab, I1, I2, I3, ref x))
                                        sprincipal[j, 0] = x;
                                    else
                                        Console.WriteLine("Failed to determine S1.");
                                }
							}
                        }
                        for (j = 0; j < 3; j++) {
                            writer.Write("SCALARS ");
                            writer.Write(DataTab[i].StepN + 1);
                            writer.Write("_");
                            writer.Write(DataTab[i].StepT.ToString(NFI));
                            writer.Write("_");
                            writer.Write(DataTab[i].Name);
                            writer.Write("_S");
                            writer.Write(j + 1);
                            writer.WriteLine(" double");
                            writer.WriteLine("LOOKUP_TABLE default");
                            k = -1;
                            foreach (KeyValuePair<int, tNode> KVP in NodeTab) {
                                if (NodeTab[KVP.Key].IsFirst) {
                                    k++;
                                    tmpstr = "";
                                    tmpstr += string.Format(NFI, "{0, 16:0.0000000E+000}", sprincipal[k,j]);
                                    writer.WriteLine(tmpstr);
                                }
                            }
                        }
                    }
                }


				// Huber-Mises-Hencky equivalent stress

				double sxxyy2, syyzz2, szzxx2, sxy2, syz2, szx2, dblval;

				for (i = 0; i < DataTab.Count; i++) {
                    if (DataTab[i].Name == "STRESS" || DataTab[i].Name == "STRESSR") {
                        writer.Write("SCALARS ");
                        writer.Write(DataTab[i].StepN + 1);
                        writer.Write("_");
                        writer.Write(DataTab[i].StepT.ToString(NFI));
                        writer.Write("_");
                        writer.Write(DataTab[i].Name);
                        writer.Write("_VMISES");
                        writer.WriteLine(" double");
                        writer.WriteLine("LOOKUP_TABLE default");
                        foreach (KeyValuePair<int, tNode> KVP in NodeTab) {
                            if (NodeTab[KVP.Key].IsFirst) {
                                if (DataTab[i].DICT.ContainsKey(KVP.Key)) {

                                    sxx = DataTab[i].DICT[KVP.Key].Data[0];
                                    syy = DataTab[i].DICT[KVP.Key].Data[1];
                                    szz = DataTab[i].DICT[KVP.Key].Data[2];
                                    sxy = DataTab[i].DICT[KVP.Key].Data[3];
                                    syz = DataTab[i].DICT[KVP.Key].Data[4];
                                    szx = DataTab[i].DICT[KVP.Key].Data[5];

                                    sxxyy2 = (sxx - syy) * (sxx - syy);
                                    syyzz2 = (syy - szz) * (syy - szz);
                                    szzxx2 = (szz - sxx) * (szz - sxx);

                                    sxy2 = sxy * sxy;
                                    syz2 = syz * syz;
                                    szx2 = szx * szx;

                                    dblval = Math.Sqrt(0.5 * (sxxyy2 + syyzz2 + szzxx2 + 6 * (sxy2 + syz2 + szx2)));
                                }
                                else
                                    dblval = 0;
                                tmpstr  = string.Format(NFI, "{0, 16:0.0000000E+000}", dblval);
								writer.WriteLine (tmpstr);
							}
						}
					}

				} // for H-M-H

			} //using
		} //method
	} //class
} //namespace

