/*

Maciek Hawryłkiewicz (06.2015)
mamah(at)wp.pl


 */

using System;
using System.IO;
using System.Reflection;
using System.Collections.Generic;
using System.Text;

namespace mmh_ccx2vtk
{
	public partial class Program
	{
		//culture info
		private static System.Globalization.NumberFormatInfo NFI;

		//input file full name
		private static string FrdFileName = "";

		private static int StepNumber = -1;
        private static double StepTime = 0;

		//nodes block
		private static int NumberOfNodes = 0;
		private static int NodeLineCounter = 0;

		//elements blok
		private static bool ELEM_FIRST_LINE = true;
		private static int NumberOfElems = 0;
		private static int CurElemType = 0;
		private static int ElemLineCounter = -1;	//starts from 0 to NumberOfElems - 1

		//nodal data block
		private static int DataBkInd = -1;
		private static int DataLineCounter = 0;

		private static bool NameOfInputFile(ref string mPlik)
		{
			string mpath = "";
			string mext = "";
			//check if file name has extension
			mext = Path.GetExtension(mPlik);
			if (string.IsNullOrEmpty(mext)) {
				//add proper extension if it was not given by user
				mPlik += ".frd";
			} else if (mext != ".frd") {
				//given improper extension 
				Console.WriteLine(ERTXT0001, mext);
				Console.Write(TXT9999);
				Console.ReadLine();
				return false;
			}
			//read path if it was given:
			mpath = Path.GetDirectoryName(mPlik);
			//if there is no path in argument then use assembly execution path:
			if (string.IsNullOrEmpty(mpath))
				mpath = Path.GetDirectoryName(Assembly.GetExecutingAssembly().CodeBase);
			//full path creation
			mPlik = mpath + Path.DirectorySeparatorChar + Path.GetFileNameWithoutExtension(mPlik) + mext;
			//removing prefix "file:" Unix / "file:\" Windows
			if (mPlik.Contains("file:")) {
				if (System.Environment.OSVersion.VersionString.Contains("Unix")) {				
					mPlik = mPlik.Substring(5);
				} else {
					mPlik = mPlik.Substring(6);
				}
			}

			Console.WriteLine(TXT0003, mPlik);

			return (File.Exists(mPlik));
		}

		private static bool ProcessNodeHeadLine(ref string mLinia, ref int NrL) {
			string[] Tstr = mLinia.Split(new Char[] {' ', '\t'}, StringSplitOptions.RemoveEmptyEntries);
			//number of nodes:
			try {
				NumberOfNodes = Convert.ToInt32(Tstr[1]);
			} catch {
				Console.WriteLine(ERTXT0010, Tstr[1], NrL);
				return false;
			}
			//set node database
			NodeTab = new System.Collections.Generic.Dictionary<int, tNode>(NumberOfNodes);
			return true;
		}

		private static bool ProcessNodeLine(ref string mLinia, ref int NrL, ref bool KONIEC)
		{
			tNode mNode = default(tNode);
			int tmpNR = 0;

			if (mLinia.IndexOf("-1") == 1) {
				try {
					NodeLineCounter += 1;
					tmpNR = Convert.ToInt32(mLinia.Substring(3, 10), NFI);
					mNode.X = Convert.ToDouble(mLinia.Substring(13, 12), NFI);
					mNode.Y = Convert.ToDouble(mLinia.Substring(25, 12), NFI);
					mNode.Z = Convert.ToDouble(mLinia.Substring(37, 12), NFI);
					mNode.IsFirst = true;
					NodeTab.Add(tmpNR, mNode);
				} catch {
					Console.WriteLine(ERTXT0011, NrL);
					return false;
				}
			} else if (mLinia.IndexOf("-3") == 1) {
				//end of reading
				if (NodeTab.Count == NodeLineCounter) {
					//read number is correct
					KONIEC = true;
					Console.WriteLine(TXT0004, NodeLineCounter);
				} else {
					//the numbers differs
					Console.WriteLine(ERTXT0012, NodeLineCounter, NodeTab.Count);
					return false;
				}
			} else {
				Console.WriteLine(ERTXT0013, NrL);
				return false;
			}

			return true;
		}

		private static int NumOfElemNodes(int ElementType) {

			switch (ElementType) {
			case 11:
				return 2;
			case 12:
			case 7:
				return 3;
			case 8:
				return 6;
			case 9:
				return 4;
			case 10:
				return 8;
			case 3:
				return 4;
			case 6:
				return 10;
			case 1:
				return 8;
			case 4:
				return 20;
			case 2:
				return 6;
			case 5:
				return 15;
			default:
				return 0;
			}

		}

		private static bool ReadNodesListLine(ref string[] tmpstr, int NrL, int ExpNum) {
			int tmpNr;
			if (tmpstr.Length == ExpNum + 1) {
				for (int i = 0; i < ExpNum; i++) {
					try {
						tmpNr = Convert.ToInt32 (tmpstr [i + 1]);
						ElementTab [ElemLineCounter].N [i] = tmpNr;
					} catch {
						Console.WriteLine(ERTXT0014, NrL.ToString());
						return false;
					}
				}
			} else {
				Console.WriteLine(ERTXT0025, NrL.ToString());
				return false;
			} 
			return true;
		}

		private static bool ProcessElementHeadLine(ref string MLinia, ref int NrL) {
			string[] TStr = MLinia.Split(new Char[] {' ', '\t'}, StringSplitOptions.RemoveEmptyEntries);
			try {
				NumberOfElems = Convert.ToInt32(TStr[1]);
			} catch {
				Console.WriteLine(ERTXT0020, TStr[1], NrL.ToString());
				return false;
			}
			ElementTab = new tElement[NumberOfElems] ;
			return true;
		}

		private static bool ProcessElementLine(ref string MLinia, ref int NrL, ref bool KONIEC)	{

			string[] TStr = null;
			int tmpNR = 0;
			int i;

			if (MLinia.IndexOf("-1") == 1) {

				ElemLineCounter++;
				TStr = MLinia.Split(new Char[] {' ', '\t'}, StringSplitOptions.RemoveEmptyEntries);

				ElementTab[ElemLineCounter].NR = Convert.ToInt32(TStr[1]);
				CurElemType = Convert.ToInt32(TStr[2]);
				ElementTab[ElemLineCounter].TYP = CurElemType;

				ElementTab [ElemLineCounter].N = new int[ NumOfElemNodes (CurElemType) ];

			} else if (MLinia.IndexOf("-2") == 1) {
				TStr = MLinia.Split(new Char[] {' ', '\t'}, StringSplitOptions.RemoveEmptyEntries);

				switch (CurElemType) {
				case 11:
					if (!ReadNodesListLine(ref TStr,NrL,2)) return false;
					break;
				case 12:
				case 7:
					if (!ReadNodesListLine(ref TStr,NrL,3)) return false;
					break;
				case 2:
				case 8:
					if (!ReadNodesListLine(ref TStr,NrL,6)) return false;
					break;
				case 3:
				case 9:
					if (!ReadNodesListLine(ref TStr,NrL,4)) return false;
					break;
				case 1:
				case 10:
					if (!ReadNodesListLine(ref TStr,NrL,8)) return false;
					break;
				case 6:
					if (!ReadNodesListLine(ref TStr,NrL,10)) return false;
					break;
				case 4:
					// two lines with ten nodes 
					if (TStr.Length == 11) {
						int offset = (ELEM_FIRST_LINE ? 0 : 10);
						for (i = 0; i < 10; i++) {
							try {
								tmpNR = Convert.ToInt32 (TStr [i + 1]);
								ElementTab [ElemLineCounter].N [i + offset] = tmpNR;

							} catch {
								Console.WriteLine (ERTXT0021, NrL.ToString ()); 
								return false;
							}
						}
					} else {
						Console.WriteLine (ERTXT0025, NrL.ToString ());
						return false;
					}
					ELEM_FIRST_LINE = !ELEM_FIRST_LINE;
					break;
				case 5:
					// two line 10 + 5 nodes - element not supported in VTK - will be processed as element type 2 (6 in VTK)
					if (ELEM_FIRST_LINE) {
						if (TStr.Length == 11) {

							for ( i = 0; i < 10; i++ ) {
								try {
									tmpNR = Convert.ToInt32(TStr[i + 1]);
									ElementTab[ElemLineCounter].N[i] = tmpNR;
								} catch {
									Console.WriteLine(ERTXT0021, NrL.ToString());
									return false;
								}
							}
						} else {
							Console.WriteLine(ERTXT0025, NrL.ToString());
							return false;
						}
						ELEM_FIRST_LINE = false;
					} else {
						//last five nodes
						if (TStr.Length == 6)
							for ( i = 0; i < 5; i++ ) 
								try {
									tmpNR = Convert.ToInt32(TStr[i + 1]);
									ElementTab[ElemLineCounter].N[i + 10] = tmpNR;
								} catch {
									Console.WriteLine(ERTXT0021, NrL.ToString());
									return false;
								}
						else {
							Console.WriteLine(ERTXT0025, NrL.ToString());
							return false;
						}
						// only first six nodes 0..5 will be transfered to VTK
						tNode mNode;
						for (i = 6; i < 15; i++ ) {
							mNode = NodeTab[ElementTab[ElemLineCounter].N[i]];
							mNode.IsFirst = false;
							NodeTab [ElementTab [ElemLineCounter].N [i]] = mNode;
						}

						ELEM_FIRST_LINE = true;
					}
					break;
				default: {
					Console.WriteLine(ERTXT0022, NrL.ToString());
					return false;
					}

				}

			} else if (MLinia.IndexOf("-3") == 1) {
				//end of elements block
				if (ElemLineCounter == NumberOfElems - 1) {
					KONIEC = true;
					Console.WriteLine(TXT0005, NumberOfElems.ToString());
				} else {
					Console.WriteLine(ERTXT0023, NumberOfElems, ElemLineCounter);
					return false;
				}
			} else {
				//improper line starting code
				Console.WriteLine(ERTXT0024, NrL.ToString());
				return false;
			}

			return true;
		}
	
		private static bool ProcessDataLine(ref string MLinia, ref int NrL, ref bool KONIEC) {
			tNodResBlRec mNodResBlRec = new tNodResBlRec();
			int tmpNr, k = 13;
			if (MLinia.IndexOf ("-1") == 1) {
				DataLineCounter++;
				mNodResBlRec.Data = new double[DataTab[DataBkInd].NumOfItems];
				try {
					tmpNr = Convert.ToInt32(MLinia.Substring(3, 10));
					for ( int i = 0; i < DataTab[DataBkInd].NumOfItems; i++ ) {
						mNodResBlRec.Data[i] = Convert.ToDouble(MLinia.Substring(k, 12), NFI);
						k += 12;
					}
					DataTab[DataBkInd].DICT.Add(tmpNr, mNodResBlRec);
				} catch {
					Console.WriteLine(ERTXT0033, NrL.ToString());
					return false;
				}
			} else if (MLinia.IndexOf("-3") == 1) {
				if (DataTab [DataBkInd].Count == DataLineCounter) {
					KONIEC = true;
					Console.WriteLine (TXT0008, DataLineCounter.ToString(), DataTab[DataBkInd].Name);
					DataLineCounter = 0;
				} else {
					Console.WriteLine (ERTXT0034, DataLineCounter, DataTab[DataBkInd].Count);
					return false;
				}
			}
			return true;
		}

		private static bool ReadFrdData()
		{
			bool CZYT_WEZLY = false;
			bool CZYT_ELEMENTY = false;
			bool CZYT_DANE = false;

			bool OK_WEZLY = false;
			bool OK_ELEMENTY = false;

			bool OK_BLOK = false;

			int NrL = 0;
			string mLinia = "";

			//linia "100C"
			int NumOfNodeData = 0;
			//linia "-4"
			string DataName = "";
			int NumOfEnt = 0;

			//linia "-5"
			int EntNamesInd = -1;

			tNodResBlock mNodResBlock;

			DataTab = new List<tNodResBlock> ();

			using (StreamReader reader = File.OpenText(FrdFileName)) {

				while (reader.Peek() >= 0) {
					mLinia = reader.ReadLine();	NrL++;

					//FLAGS
					if (mLinia.IndexOf ("2C") == 4) {
						if ( ProcessNodeHeadLine(ref mLinia, ref NrL) ) {
							CZYT_WEZLY = true;
							continue;
						} else 
							break;
					}

					if (mLinia.IndexOf ("3C") == 4) {
						if ( ProcessElementHeadLine (ref mLinia, ref NrL) ) {
							CZYT_ELEMENTY = true;
							continue;
						} else 
							break;
					}

//					//Parameter Header Record:
//					if (mLinia.IndexOf ("    1P") == 0)
//						continue;

					//Nodal Result Block:
					if (mLinia.IndexOf("100C") == 2) {
						/*
						Format:(1X,’ 100’,’C’,6A1,E12.5,I12,20A1,I2,I5,10A1,I2)
						Values: KEY,CODE,SETNAME,VALUE,NUMNOD,TEXT,ICTYPE,NUMSTP,ANALYS,FORMAT
						 */
						try {

                            StepTime = Convert.ToDouble(mLinia.Substring(12, 12), NFI);
                            NumOfNodeData = Convert.ToInt32(mLinia.Substring(24, 12));
                            int SNr = Convert.ToInt32(mLinia.Substring(58, 5)) - 1;

                            if (StepNumber != SNr) {
								StepNumber = SNr;
                                Console.WriteLine(TXT0006, StepNumber + 1);
                                Console.WriteLine(mLinia.Substring(12, 12));
                                Console.WriteLine(StepTime);
                            }
						} catch {
							Console.WriteLine (ERTXT0030, NrL);
							return false;
						}
						continue;
					}

					/*
							Format:(1X,I2,2X,8A1,2I5)
							Values: KEY, NAME, NCOMPS, IRTYPE
							Where: KEY = -4
							NAME = Dataset name to be used in the menu
							NCOMPS = Number of entities
							IRTYPE = 1 Nodal data, material independent
							2 Nodal data, material dependant
							3 Element data at nodes (not used)
					 */
					if (mLinia.IndexOf ("-4") == 1) {
						try {
//							DataName = mLinia.Substring(5, 8).Trim();
							DataName = mLinia.Substring(5, 8).Replace(" ", "");
							NumOfEnt = Convert.ToInt32 (mLinia.Substring(13, 5));
						} catch {
							Console.WriteLine (ERTXT0031, NrL);
							return false;
						}


						Console.WriteLine(TXT0007, DataName, NumOfNodeData, NumOfEnt);

						mNodResBlock = new tNodResBlock (DataName, NumOfNodeData, NumOfEnt, StepNumber, StepTime);
						DataTab.Add(mNodResBlock);
						DataBkInd++; //current DataTab index (current block of data index)
						EntNamesInd = -1; //prepare to read items from lines code '-5'
						continue;
					}

					if (mLinia.IndexOf ("-5") == 1) {
						EntNamesInd++;
						try {
//							DataTab[DataBkInd].ItemNameTab[EntNamesInd] = mLinia.Substring(5,8).Trim();
							DataTab[DataBkInd].ItemNameTab[EntNamesInd] = mLinia.Substring(5,8).Replace(" ", "");
						} catch {
							Console.WriteLine (ERTXT0032, NrL);
							return false;
						}

						// if there is an item to be auto calculated by cgx 
						// reduce number of expected values in data lines
						try
                        {
                            if (mLinia.IndexOf(" 1ALL") > -1 ) {
								mNodResBlock = DataTab[DataBkInd];
								mNodResBlock.NumOfItems--;
								DataTab[DataBkInd] = mNodResBlock;
							}
						} catch {/*nothing to do*/}	

						//check if this is the last code '-5' line
						if (DataTab [DataBkInd].DeclNumOfItems == EntNamesInd + 1)
							CZYT_DANE = true;

						continue;
					}

					//Processing
					if (CZYT_WEZLY) {
						if (ProcessNodeLine (ref mLinia, ref NrL, ref OK_WEZLY)) {
							if (OK_WEZLY)
								CZYT_WEZLY = false;
						} else 
							break;

					} else if (OK_WEZLY && CZYT_ELEMENTY) {
						if (ProcessElementLine (ref mLinia, ref NrL, ref OK_ELEMENTY)) {
							if (OK_ELEMENTY)
								CZYT_ELEMENTY = false;
						} else 
							break;

					} else if (OK_WEZLY && OK_ELEMENTY && CZYT_DANE) {
						if (ProcessDataLine (ref mLinia, ref NrL, ref OK_BLOK)) {
							if (OK_BLOK) {
								OK_BLOK = false;
								CZYT_DANE = false;
								DataLineCounter = 0;
							}
						} else
							break;
					} //else
				} //while
			} //using
			//	nodal result datas are not required, the model is enough
			return (OK_WEZLY & OK_ELEMENTY);
		}

		public static void Main(string[] args)
		{

			NFI = new System.Globalization.NumberFormatInfo ();
			NFI.NumberDecimalSeparator = ".";

			Console.WriteLine (TXT0000);
			Console.WriteLine (TXT0001 + System.Environment.OSVersion.VersionString);

			//for debugging
			//args = new string[] { "/media/mmh/E_Windows/_CALCULIX_BCONV_WORKING/dupa/dupa.frd" };

			if (args.Length != 1) {
				Console.WriteLine(ERTXT0000);
				PrintHelp ();
			} else {

				Console.WriteLine (TXT0002 + args[0]);
				FrdFileName = args[0];

				if (NameOfInputFile(ref FrdFileName)) {
					if (ReadFrdData()) {
						Console.WriteLine (TXT0009, StepNumber + 1);
						MakeVTKFile ();
					}
				} else {
					Console.WriteLine(ERTXT0002);
					PrintHelp ();
				}
			}

			Console.WriteLine();
			Console.Write(TXT9999);
			//Console.ReadLine();

		}
	}
}
